/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.vew.engvew.engstelle;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dambach.dav.daf.EmpfaengerMitAnAbmeldung;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Modelliert den Datenverteiler-Typ typ.situation mit seinen Parametern und Eigenschaften
 * 
 * @author Dambach Werke GmbH
 * @author Stefan Sans
 * @version $Revision: 1.1 $ / $Date: 2008/10/31 13:14:01 $ / ($Author: Sans $)
 */
public class Situation implements EmpfaengerMitAnAbmeldung
{
  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  /**
   * Systemobjekt vom Typ Situation
   */
  private SystemObject _davObjekt;

  /**
   * Datenverteiler-Verbindung
   */
  private ClientDavInterface _connection;

  /**
   * Konstruiert eine Instanz vom Typ Situation
   *
   * @param connection Datenverteiler-Verbindung
   * @param davObjekt Systemobjekt vom Typ Situation
   */
  protected Situation(ClientDavInterface connection, SystemObject davObjekt)
  {
    _connection = connection;
    _davObjekt = davObjekt;
  }

  @Override
  public String toString()
  {
    return getObjekt() != null ? getObjekt().getName() : "";
  }

  /**
   * Lage dieser Situation
   */
  private Lage _lage;

  /**
   * Setze neue Lage für diese Situation
   *
   * @param lage neue Lage
   */
  protected void setLage(Lage lage)
  {
    _lage = lage;
  }

  /**
   * Bestimme Lage für diese Situation
   *
   * @return aktuelle Lage
   */
  public Lage getLage()
  {
    return _lage;
  }

  /**
   * Modelliert die Lage einer Situation
   *
   * @author Stefan Sans
   */
  public static class Lage
  {
    /**
     * Die Strassensegmente, die die Lage beschreiben
     */
    public final Array strassenSegmente;

    /**
     * Offset am Anfang der ersten Strassensegments
     */
    public final float offsetVor;

    /**
     * Offset am Ende des letzten Strassensegments
     */
    public final float offsetNach;

    /**
     * Erzeuge neue Lage
     * @param strassenSegmente die Strassensegmente, die die Lage beschreiben
     * @param offsetVor Offset am Anfang der ersten Strassensegments
     * @param offsetNach Offset am Ende des letzten Strassensegments
     */
    public Lage(Array strassenSegmente, float offsetVor, float offsetNach)
    {
      this.strassenSegmente = strassenSegmente;
      this.offsetVor = offsetVor;
      this.offsetNach = offsetNach;
    }

  }

  /**
   * Gültigkeit dieser Situation
   */
  private Gueltigkeit _gueltigkeit;

  /**
   * Setze neue Gültikeit für diese Situation
   *
   * @param gueltigkeit neue Gültigkeit
   */
  protected void setGueltigkeit(Gueltigkeit gueltigkeit)
  {
    _gueltigkeit = gueltigkeit;
  }

  /**
   * Bestimme Gültigkeit für diese Situation
   *
   * @return aktuelle Gültigkeit
   */
  public Gueltigkeit getGueltigkeit()
  {
    return _gueltigkeit;
  }

  /**
   * Modelliert die Gültigkeit einer Situation
   *
   * @author Stefan Sans
   */
  public static class Gueltigkeit
  {
    /**
     * Anfang der Gültigkeit in ms
     */
    public final long beginn;

    /**
     * Dauer der Gültigkeit in ms
     */
    public final long dauer;

    /**
     * Erzeuge neue Gültigkeit
     *
     * @param beginn Anfang der Gültigkeit in ms
     * @param dauer Dauer der Gültigkeit in ms
     */
    public Gueltigkeit(long beginn, long dauer)
    {
      this.beginn = beginn;
      this.dauer = dauer;
    }

  }

  /**
   * Prüft, ob diese Instanz zum Zeitpunkt zeitpunkt gültig ist
   *
   * @param zeitpunkt der Zeitpunkt, für den geprüft werden soll
   * @return true, falls diese Instanz gültig ist, false sonst
   */
  public boolean pruefeGueltigkeit(long zeitpunkt)
  {
    if (getGueltigkeit() == null)
    {
      _debug.fine("Gültigkeit ist nicht spezifiziert für " + this);
      return false;
    }

    if (getGueltigkeit().beginn > zeitpunkt || // Engstelle beginnt in der Zukunft
        (getGueltigkeit().dauer > 0 && // Engstelle ist bereits beendet - abgelaufen
        getGueltigkeit().beginn + getGueltigkeit().dauer < zeitpunkt))
    {
      _debug.fine(this + " ist aktuell nicht gültig");
      return false;
    }

    return true;
  }

  /**
   * Prüft, ob die Lage dieser Instanz ausreichend spezifiziert ist
   *
   * @return true, falls die LAge ausreichend spezifiziert ist, false sonst
   */
  public boolean pruefeLage()
  {
    if (getLage() == null)
    {
      _debug.fine("Lage ist nicht spezifiziert für " + this);
      return false;
    }

    if (getLage().strassenSegmente.getLength() == 0)
    {
      _debug.fine("Lage nicht ausreichend spezifiziert für " +  this +
          ": keine Strassensegmente");
      return false;
    }

    return true;
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.vew.vewallg.empf.EmpfaengerMitAnmeldung#getConnection()
   */
  public ClientDavInterface getConnection()
  {
    return _connection;
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.vew.vewallg.empf.EmpfaengerMitAnmeldung#getObjekt()
   */
  public SystemObject getObjekt()
  {
    return _davObjekt;
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.vew.vewallg.empf.EmpfaengerMitAnmeldung#anmelden()
   */
  public void anmelden()
  {
    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.situationsEigenschaften");
    Aspect asp = getConnection().getDataModel().getAspect("asp.parameterSoll");
    DataDescription datenbeschreibung = new DataDescription(atg, asp);

    getConnection().subscribeReceiver(this, getObjekt(), datenbeschreibung,
        ReceiveOptions.normal(), ReceiverRole.receiver());
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.vew.vewallg.empf.EmpfaengerMitAnAbmeldung#abmelden()
   */
  public void abmelden()
  {
    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.situationsEigenschaften");
    Aspect asp = getConnection().getDataModel().getAspect("asp.parameterSoll");
    DataDescription datenbeschreibung = new DataDescription(atg, asp);

    getConnection().unsubscribeReceiver(this, getObjekt(), datenbeschreibung);
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
   */
  public synchronized void update(ResultData[] resultData)
  {
    // Ausgaben
    for (ResultData data : resultData)
    {
      Aspect asp = data.getDataDescription().getAspect();
      AttributeGroup atg = data.getDataDescription().getAttributeGroup();
      short sv = data.getDataDescription().getSimulationVariant();

      _debug.finest("Daten empfangen mit " + atg.getPid() + "/" +
          asp.getPid() + "/" + sv  + ":\n" + data);

      if (asp.getPid().equals("asp.parameterSoll") &&
          atg.getPid().equals("atg.situationsEigenschaften"))
      {
        if (data.hasData())
        {
          _debug.finer("Bearbeite Daten mit " + atg.getPid() + "/" +
              asp.getPid() + "/" + sv + ":\n" + data);

          Data item = data.getData().getItem("StartZeit");
          try
          {
            long beginn = item.asTimeValue().getMillis();
            item = data.getData().getItem("Dauer");
            long dauer = item.asTimeValue().getMillis();
            setGueltigkeit(new Gueltigkeit(beginn, dauer));
            _debug.config("Neue Gültigkeit für " + getObjekt() + ": " + beginn + ", " + dauer);
          }
          catch(RuntimeException e)
          {
            _debug.warning(getObjekt() + " ist ohne Gültigkeit", e);
          }

          try
          {
            item = data.getData().getItem("StraßenSegment");
            Array strassenSegmente = item.asArray();
            if (strassenSegmente != null && strassenSegmente.getLength() > 0)
            {
              item = data.getData().getItem("StartOffset");
              float startOffset = item.asScaledValue().floatValue();
              item = data.getData().getItem("EndOffset");
              float endOffset = item.asScaledValue().floatValue();
              setLage(new Lage(strassenSegmente, startOffset, endOffset));
              _debug.config("Neue Lage für " + getObjekt() + " - erstreckt sich über " +
                  strassenSegmente.getLength() + " Strassensegment(e)");
            }
            else
            {
              setLage(null);
            }
          }  
          catch(RuntimeException e)
          {
            setLage(null);
            _debug.warning(getObjekt() + " ist ohne Lage", e);
          }
        }
        else
        {
          setGueltigkeit(null);
          setLage(null);
        }
      }
    }
  }  
  
}
